from typing import Any, Optional

class Deleted:
    key: Any
    version_id: Any
    delete_marker: Any
    delete_marker_version_id: Any
    def __init__(self, key: Optional[Any] = ..., version_id: Optional[Any] = ..., delete_marker: bool = ..., delete_marker_version_id: Optional[Any] = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...

class Error:
    key: Any
    version_id: Any
    code: Any
    message: Any
    def __init__(self, key: Optional[Any] = ..., version_id: Optional[Any] = ..., code: Optional[Any] = ..., message: Optional[Any] = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...

class MultiDeleteResult:
    bucket: Any
    deleted: Any
    errors: Any
    def __init__(self, bucket: Optional[Any] = ...) -> None: ...
    def startElement(self, name, attrs, connection): ...
    def endElement(self, name, value, connection): ...



import sys
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Text, Tuple

Event = NamedTuple('Event', [
    ('time', float),
    ('priority', Any),
    ('action', Callable[..., Any]),
    ('argument', Tuple[Any, ...]),
    ('kwargs', Dict[Text, Any]),
])

class scheduler:
    if sys.version_info >= (3, 3):
        def __init__(self, timefunc: Callable[[], float] = ..., delayfunc: Callable[[float], None] = ...) -> None: ...
        def enterabs(self, time: float, priority: Any, action: Callable[..., Any], argument: Tuple[Any, ...] = ..., kwargs: Dict[str, Any] = ...) -> Event: ...
        def enter(self, delay: float, priority: Any, action: Callable[..., Any], argument: Tuple[Any, ...] = ..., kwargs: Dict[str, Any] = ...) -> Event: ...
        def run(self, blocking: bool = ...) -> Optional[float]: ...
    else:
        def __init__(self, timefunc: Callable[[], float], delayfunc: Callable[[float], None]) -> None: ...
        def enterabs(self, time: float, priority: Any, action: Callable[..., Any], argument: Tuple[Any, ...]) -> Event: ...
        def enter(self, delay: float, priority: Any, action: Callable[..., Any], argument: Tuple[Any, ...]) -> Event: ...
        def run(self) -> None: ...

    def cancel(self, event: Event) -> None: ...
    def empty(self) -> bool: ...
    @property
    def queue(self) -> List[Event]: ...



"""
Wrapper around the eventloop that gives some time to the Tkinter GUI to process
events when it's loaded and while we are waiting for input at the REPL. This
way we don't block the UI of for instance ``turtle`` and other Tk libraries.

(Normally Tkinter registeres it's callbacks in ``PyOS_InputHook`` to integrate
in readline. ``prompt-toolkit`` doesn't understand that input hook, but this
will fix it for Tk.)
"""
from rp.prompt_toolkit.shortcuts import create_eventloop as _create_eventloop
import sys
import time

__all__ = (
    'create_eventloop',
)


def _inputhook_tk(inputhook_context):
    """
    Inputhook for Tk.
    Run the Tk eventloop until prompt-toolkit needs to process the next input.
    """
    # Get the current TK application.
    import _tkinter  # Keep this imports inline!
    from six.moves import tkinter
    root = tkinter._default_root

    def wait_using_filehandler():
        """
        Run the TK eventloop until the file handler that we got from the
        inputhook becomes readable.
        """
        # Add a handler that sets the stop flag when `prompt-toolkit` has input
        # to process.
        stop = [False]
        def done(*a):
            stop[0] = True

        root.createfilehandler(inputhook_context.fileno(), _tkinter.READABLE, done)

        # Run the TK event loop as long as we don't receive input.
        while root.dooneevent(_tkinter.ALL_EVENTS):
            if stop[0]:
                break

        root.deletefilehandler(inputhook_context.fileno())

    def wait_using_polling():
        """
        Windows TK doesn't support 'createfilehandler'.
        So, run the TK eventloop and poll until input is ready.
        """
        while not inputhook_context.input_is_ready():
            while root.dooneevent(_tkinter.ALL_EVENTS | _tkinter.DONT_WAIT):
                 pass
            # Sleep to make the CPU idle, but not too long, so that the UI
            # stays responsive.
            time.sleep(.01)

    if root is not None:
        if hasattr(root, 'createfilehandler'):
            wait_using_filehandler()
        else:
            wait_using_polling()


def _inputhook(inputhook_context):
    # Only call the real input hook when the 'Tkinter' library was loaded.
    if 'Tkinter' in sys.modules or 'tkinter' in sys.modules:
        _inputhook_tk(inputhook_context)


def create_eventloop():
    return _create_eventloop(inputhook=_inputhook)



SERVER_STATUS_IN_TRANS: int
SERVER_STATUS_AUTOCOMMIT: int
SERVER_MORE_RESULTS_EXISTS: int
SERVER_QUERY_NO_GOOD_INDEX_USED: int
SERVER_QUERY_NO_INDEX_USED: int
SERVER_STATUS_CURSOR_EXISTS: int
SERVER_STATUS_LAST_ROW_SENT: int
SERVER_STATUS_DB_DROPPED: int
SERVER_STATUS_NO_BACKSLASH_ESCAPES: int
SERVER_STATUS_METADATA_CHANGED: int



# Stubs for distutils.text_file

from typing import IO, List, Optional, Tuple, Union

class TextFile:
    def __init__(self, filename: Optional[str] = ...,
                 file: Optional[IO[str]] = ...,
                 *, strip_comments: bool = ...,
                 lstrip_ws: bool = ..., rstrip_ws: bool = ...,
                 skip_blanks: bool = ..., join_lines: bool = ...,
                 collapse_join: bool = ...) -> None: ...
    def open(self, filename: str) -> None: ...
    def close(self) -> None: ...
    def warn(self, msg: str,
             line: Union[List[int], Tuple[int, int], int] = ...) -> None: ...
    def readline(self) -> Optional[str]: ...
    def readlines(self) -> List[str]: ...
    def unreadline(self, line: str) -> str: ...



#Display green rainfall of characters like outta da matrix!
#*Puts on hacker goggles*
from rp import *
pip_import('unimatrix','git+https://github.com/will8211/unimatrix.git')
import unimatrix
unimatrix.main()



from typing import Any

def ReadTag(buffer, pos): ...
def EnumDecoder(field_number, is_repeated, is_packed, key, new_default): ...

Int32Decoder: Any
Int64Decoder: Any
UInt32Decoder: Any
UInt64Decoder: Any
SInt32Decoder: Any
SInt64Decoder: Any
Fixed32Decoder: Any
Fixed64Decoder: Any
SFixed32Decoder: Any
SFixed64Decoder: Any
FloatDecoder: Any
DoubleDecoder: Any
BoolDecoder: Any

def StringDecoder(field_number, is_repeated, is_packed, key, new_default): ...
def BytesDecoder(field_number, is_repeated, is_packed, key, new_default): ...
def GroupDecoder(field_number, is_repeated, is_packed, key, new_default): ...
def MessageDecoder(field_number, is_repeated, is_packed, key, new_default): ...

MESSAGE_SET_ITEM_TAG: Any

def MessageSetItemDecoder(extensions_by_number): ...
def MapDecoder(field_descriptor, new_default, is_message_map): ...

SkipField: Any



"""
To ensure compatibility from Python ``2.7`` - ``3.x``, a module has been
created. Clearly there is huge need to use conforming syntax.
"""
from __future__ import print_function
import atexit
import errno
import functools
import sys
import os
import re
import pkgutil
import warnings
import inspect
import subprocess
import weakref
try:
    import importlib
except ImportError:
    pass
from zipimport import zipimporter

from rp.libs.jedi.file_io import KnownContentFileIO, ZipFileIO

is_py3 = sys.version_info[0] >= 3
is_py35 = is_py3 and sys.version_info[1] >= 5
py_version = int(str(sys.version_info[0]) + str(sys.version_info[1]))


class DummyFile(object):
    def __init__(self, loader, string):
        self.loader = loader
        self.string = string

    def read(self):
        return self.loader.get_source(self.string)

    def close(self):
        del self.loader


def find_module_py34(string, path=None, full_name=None, is_global_search=True):
    spec = None
    loader = None

    for finder in sys.meta_path:
        if is_global_search and finder != importlib.machinery.PathFinder:
            p = None
        else:
            p = path
        try:
            find_spec = finder.find_spec
        except AttributeError:
            # These are old-school clases that still have a different API, just
            # ignore those.
            continue

        spec = find_spec(string, p)
        if spec is not None:
            loader = spec.loader
            if loader is None and not spec.has_location:
                # This is a namespace package.
                full_name = string if not path else full_name
                implicit_ns_info = ImplicitNSInfo(full_name, spec.submodule_search_locations._path)
                return implicit_ns_info, True
            break

    return find_module_py33(string, path, loader)


def find_module_py33(string, path=None, loader=None, full_name=None, is_global_search=True):
    loader = loader or importlib.machinery.PathFinder.find_module(string, path)

    if loader is None and path is None:  # Fallback to find builtins
        try:
            with warnings.catch_warnings(record=True):
                # Mute "DeprecationWarning: Use importlib.util.find_spec()
                # instead." While we should replace that in the future, it's
                # probably good to wait until we deprecate Python 3.3, since
                # it was added in Python 3.4 and find_loader hasn't been
                # removed in 3.6.
                loader = importlib.find_loader(string)
        except ValueError as e:
            # See #491. Importlib might raise a ValueError, to avoid this, we
            # just raise an ImportError to fix the issue.
            raise ImportError("Originally  " + repr(e))

    if loader is None:
        raise ImportError("Couldn't find a loader for {}".format(string))

    return _from_loader(loader, string)


def _from_loader(loader, string):
    is_package = loader.is_package(string)
    try:
        get_filename = loader.get_filename
    except AttributeError:
        return None, is_package
    else:
        module_path = cast_path(get_filename(string))

    # To avoid unicode and read bytes, "overwrite" loader.get_source if
    # possible.
    f = type(loader).get_source
    if is_py3 and f is not importlib.machinery.SourceFileLoader.get_source:
        # Unfortunately we are reading unicode here, not bytes.
        # It seems hard to get bytes, because the zip importer
        # logic just unpacks the zip file and returns a file descriptor
        # that we cannot as easily access. Therefore we just read it as
        # a string in the cases where get_source was overwritten.
        code = loader.get_source(string)
    else:
        code = _get_source(loader, string)

    if code is None:
        return None, is_package
    if isinstance(loader, zipimporter):
        return ZipFileIO(module_path, code, cast_path(loader.archive)), is_package

    return KnownContentFileIO(module_path, code), is_package


def _get_source(loader, fullname):
    """
    This method is here as a replacement for SourceLoader.get_source. That
    method returns unicode, but we prefer bytes.
    """
    path = loader.get_filename(fullname)
    try:
        return loader.get_data(path)
    except OSError:
        raise ImportError('source not available through get_data()',
                          name=fullname)


def find_module_pre_py3(string, path=None, full_name=None, is_global_search=True):
    # This import is here, because in other places it will raise a
    # DeprecationWarning.
    import imp
    try:
        module_file, module_path, description = imp.find_module(string, path)
        module_type = description[2]
        is_package = module_type is imp.PKG_DIRECTORY
        if is_package:
            # In Python 2 directory package imports are returned as folder
            # paths, not __init__.py paths.
            p = os.path.join(module_path, '__init__.py')
            try:
                module_file = open(p)
                module_path = p
            except FileNotFoundError:
                pass
        elif module_type != imp.PY_SOURCE:
            if module_file is not None:
                module_file.close()
            module_file = None

        if module_file is None:
            code = None
            return None, is_package

        with module_file:
            code = module_file.read()
        return KnownContentFileIO(cast_path(module_path), code), is_package
    except ImportError:
        pass

    if path is None:
        path = sys.path
    for item in path:
        loader = pkgutil.get_importer(item)
        if loader:
            loader = loader.find_module(string)
            if loader is not None:
                return _from_loader(loader, string)
    raise ImportError("No module named {}".format(string))


find_module = find_module_py34 if is_py3 else find_module_pre_py3
find_module.__doc__ = """
Provides information about a module.

This function isolates the differences in importing libraries introduced with
python 3.3 on; it gets a module name and optionally a path. It will return a
tuple containin an open file for the module (if not builtin), the filename
or the name of the module if it is a builtin one and a boolean indicating
if the module is contained in a package.
"""


def _iter_modules(paths, prefix=''):
    # Copy of pkgutil.iter_modules adapted to work with namespaces

    for path in paths:
        importer = pkgutil.get_importer(path)

        if not isinstance(importer, importlib.machinery.FileFinder):
            # We're only modifying the case for FileFinder. All the other cases
            # still need to be checked (like zip-importing). Do this by just
            # calling the pkgutil version.
            for mod_info in pkgutil.iter_modules([path], prefix):
                yield mod_info
            continue

        # START COPY OF pkutils._iter_file_finder_modules.
        if importer.path is None or not os.path.isdir(importer.path):
            return

        yielded = {}

        try:
            filenames = os.listdir(importer.path)
        except OSError:
            # ignore unreadable directories like import does
            filenames = []
        filenames.sort()  # handle packages before same-named modules

        for fn in filenames:
            modname = inspect.getmodulename(fn)
            if modname == '__init__' or modname in yielded:
                continue

            # jedi addition: Avoid traversing special directories
            if fn.startswith('.') or fn == '__pycache__':
                continue

            path = os.path.join(importer.path, fn)
            ispkg = False

            if not modname and os.path.isdir(path) and '.' not in fn:
                modname = fn
                # A few jedi modifications: Don't check if there's an
                # __init__.py
                try:
                    os.listdir(path)
                except OSError:
                    # ignore unreadable directories like import does
                    continue
                ispkg = True

            if modname and '.' not in modname:
                yielded[modname] = 1
                yield importer, prefix + modname, ispkg
        # END COPY


iter_modules = _iter_modules if py_version >= 34 else pkgutil.iter_modules


class ImplicitNSInfo(object):
    """Stores information returned from an implicit namespace spec"""
    def __init__(self, name, paths):
        self.name = name
        self.paths = paths


if is_py3:
    all_suffixes = importlib.machinery.all_suffixes
else:
    def all_suffixes():
        # Is deprecated and raises a warning in Python 3.6.
        import imp
        return [suffix for suffix, _, _ in imp.get_suffixes()]


# unicode function
try:
    unicode = unicode
except NameError:
    unicode = str


# re-raise function
if is_py3:
    def reraise(exception, traceback):
        raise exception.with_traceback(traceback)
else:
    eval(compile("""
def reraise(exception, traceback):
    raise exception, None, traceback
""", 'blub', 'exec'))

reraise.__doc__ = """
Re-raise `exception` with a `traceback` object.

Usage::

    reraise(Exception, sys.exc_info()[2])

"""


def use_metaclass(meta, *bases):
    """ Create a class with a metaclass. """
    if not bases:
        bases = (object,)
    return meta("Py2CompatibilityMetaClass", bases, {})


try:
    encoding = sys.stdout.encoding
    if encoding is None:
        encoding = 'utf-8'
except AttributeError:
    encoding = 'ascii'


def u(string, errors='strict'):
    """Cast to unicode DAMMIT!
    Written because Python2 repr always implicitly casts to a string, so we
    have to cast back to a unicode (and we now that we always deal with valid
    unicode, because we check that in the beginning).
    """
    if isinstance(string, bytes):
        return unicode(string, encoding='UTF-8', errors=errors)
    return string


def cast_path(obj):
    """
    Take a bytes or str path and cast it to unicode.

    Apparently it is perfectly fine to pass both byte and unicode objects into
    the sys.path. This probably means that byte paths are normal at other
    places as well.

    Since this just really complicates everything and Python 2.7 will be EOL
    soon anyway, just go with always strings.
    """
    return u(obj, errors='replace')


def force_unicode(obj):
    # Intentionally don't mix those two up, because those two code paths might
    # be different in the future (maybe windows?).
    return cast_path(obj)


try:
    import builtins  # module name in python 3
except ImportError:
    import __builtin__ as builtins  # noqa: F401


import ast  # noqa: F401


def literal_eval(string):
    return ast.literal_eval(string)


try:
    from itertools import zip_longest
except ImportError:
    from itertools import izip_longest as zip_longest  # Python 2  # noqa: F401

try:
    FileNotFoundError = FileNotFoundError
except NameError:
    FileNotFoundError = IOError

try:
    IsADirectoryError = IsADirectoryError
except NameError:
    IsADirectoryError = IOError

try:
    PermissionError = PermissionError
except NameError:
    PermissionError = IOError


def no_unicode_pprint(dct):
    """
    Python 2/3 dict __repr__ may be different, because of unicode differens
    (with or without a `u` prefix). Normally in doctests we could use `pprint`
    to sort dicts and check for equality, but here we have to write a separate
    function to do that.
    """
    import pprint
    s = pprint.pformat(dct)
    print(re.sub("u'", "'", s))


def utf8_repr(func):
    """
    ``__repr__`` methods in Python 2 don't allow unicode objects to be
    returned. Therefore cast them to utf-8 bytes in this decorator.
    """
    def wrapper(self):
        result = func(self)
        if isinstance(result, unicode):
            return result.encode('utf-8')
        else:
            return result

    if is_py3:
        return func
    else:
        return wrapper


if is_py3:
    import queue
else:
    import Queue as queue  # noqa: F401

try:
    # Attempt to load the C implementation of pickle on Python 2 as it is way
    # faster.
    import cPickle as pickle
except ImportError:
    import pickle
if sys.version_info[:2] == (3, 3):
    """
    Monkeypatch the unpickler in Python 3.3. This is needed, because the
    argument `encoding='bytes'` is not supported in 3.3, but badly needed to
    communicate with Python 2.
    """

    class NewUnpickler(pickle._Unpickler):
        dispatch = dict(pickle._Unpickler.dispatch)

        def _decode_string(self, value):
            # Used to allow strings from Python 2 to be decoded either as
            # bytes or Unicode strings.  This should be used only with the
            # STRING, BINSTRING and SHORT_BINSTRING opcodes.
            if self.encoding == "bytes":
                return value
            else:
                return value.decode(self.encoding, self.errors)

        def load_string(self):
            data = self.readline()[:-1]
            # Strip outermost quotes
            if len(data) >= 2 and data[0] == data[-1] and data[0] in b'"\'':
                data = data[1:-1]
            else:
                raise pickle.UnpicklingError("the STRING opcode argument must be quoted")
            self.append(self._decode_string(pickle.codecs.escape_decode(data)[0]))
        dispatch[pickle.STRING[0]] = load_string

        def load_binstring(self):
            # Deprecated BINSTRING uses signed 32-bit length
            len, = pickle.struct.unpack('<i', self.read(4))
            if len < 0:
                raise pickle.UnpicklingError("BINSTRING pickle has negative byte count")
            data = self.read(len)
            self.append(self._decode_string(data))
        dispatch[pickle.BINSTRING[0]] = load_binstring

        def load_short_binstring(self):
            len = self.read(1)[0]
            data = self.read(len)
            self.append(self._decode_string(data))
        dispatch[pickle.SHORT_BINSTRING[0]] = load_short_binstring

    def load(file, fix_imports=True, encoding="ASCII", errors="strict"):
        return NewUnpickler(file, fix_imports=fix_imports,
                            encoding=encoding, errors=errors).load()

    def loads(s, fix_imports=True, encoding="ASCII", errors="strict"):
        if isinstance(s, str):
            raise TypeError("Can't load pickle from unicode string")
        file = pickle.io.BytesIO(s)
        return NewUnpickler(file, fix_imports=fix_imports,
                            encoding=encoding, errors=errors).load()

    pickle.Unpickler = NewUnpickler
    pickle.load = load
    pickle.loads = loads


def pickle_load(file):
    try:
        if is_py3:
            return pickle.load(file, encoding='bytes')
        return pickle.load(file)
    # Python on Windows don't throw EOF errors for pipes. So reraise them with
    # the correct type, which is caught upwards.
    except OSError:
        if sys.platform == 'win32':
            raise EOFError()
        raise


def _python2_dct_keys_to_unicode(data):
    """
    Python 2 stores object __dict__ entries as bytes, not unicode, correct it
    here. Python 2 can deal with both, Python 3 expects unicode.
    """
    if isinstance(data, tuple):
        return tuple(_python2_dct_keys_to_unicode(x) for x in data)
    elif isinstance(data, list):
        return list(_python2_dct_keys_to_unicode(x) for x in data)
    elif hasattr(data, '__dict__') and type(data.__dict__) == dict:
        data.__dict__ = {unicode(k): v for k, v in data.__dict__.items()}
    return data


def pickle_dump(data, file, protocol):
    try:
        if not is_py3:
            data = _python2_dct_keys_to_unicode(data)
        pickle.dump(data, file, protocol)
        # On Python 3.3 flush throws sometimes an error even though the writing
        # operation should be completed.
        file.flush()
    # Python on Windows don't throw EPIPE errors for pipes. So reraise them with
    # the correct type and error number.
    except OSError:
        if sys.platform == 'win32':
            raise IOError(errno.EPIPE, "Broken pipe")
        raise


# Determine the highest protocol version compatible for a given list of Python
# versions.
def highest_pickle_protocol(python_versions):
    protocol = 4
    for version in python_versions:
        if version[0] == 2:
            # The minimum protocol version for the versions of Python that we
            # support (2.7 and 3.3+) is 2.
            return 2
        if version[1] < 4:
            protocol = 3
    return protocol


try:
    from inspect import Parameter
except ImportError:
    class Parameter(object):
        POSITIONAL_ONLY = object()
        POSITIONAL_OR_KEYWORD = object()
        VAR_POSITIONAL = object()
        KEYWORD_ONLY = object()
        VAR_KEYWORD = object()


class GeneralizedPopen(subprocess.Popen):
    def __init__(self, *args, **kwargs):
        if os.name == 'nt':
            try:
                # Was introduced in Python 3.7.
                CREATE_NO_WINDOW = subprocess.CREATE_NO_WINDOW
            except AttributeError:
                CREATE_NO_WINDOW = 0x08000000
            kwargs['creationflags'] = CREATE_NO_WINDOW
        # The child process doesn't need file descriptors except 0, 1, 2.
        # This is unix only.
        kwargs['close_fds'] = 'posix' in sys.builtin_module_names
        super(GeneralizedPopen, self).__init__(*args, **kwargs)


# shutil.which is not available on Python 2.7.
def which(cmd, mode=os.F_OK | os.X_OK, path=None):
    """Given a command, mode, and a PATH string, return the path which
    conforms to the given mode on the PATH, or None if there is no such
    file.

    `mode` defaults to os.F_OK | os.X_OK. `path` defaults to the result
    of os.environ.get("PATH"), or can be overridden with a custom search
    path.

    """
    # Check that a given file can be accessed with the correct mode.
    # Additionally check that `file` is not a directory, as on Windows
    # directories pass the os.access check.
    def _access_check(fn, mode):
        return (os.path.exists(fn) and os.access(fn, mode)
                and not os.path.isdir(fn))

    # If we're given a path with a directory part, look it up directly rather
    # than referring to PATH directories. This includes checking relative to the
    # current directory, e.g. ./script
    if os.path.dirname(cmd):
        if _access_check(cmd, mode):
            return cmd
        return None

    if path is None:
        path = os.environ.get("PATH", os.defpath)
    if not path:
        return None
    path = path.split(os.pathsep)

    if sys.platform == "win32":
        # The current directory takes precedence on Windows.
        if os.curdir not in path:
            path.insert(0, os.curdir)

        # PATHEXT is necessary to check on Windows.
        pathext = os.environ.get("PATHEXT", "").split(os.pathsep)
        # See if the given file matches any of the expected path extensions.
        # This will allow us to short circuit when given "python.exe".
        # If it does match, only test that one, otherwise we have to try
        # others.
        if any(cmd.lower().endswith(ext.lower()) for ext in pathext):
            files = [cmd]
        else:
            files = [cmd + ext for ext in pathext]
    else:
        # On other platforms you don't have things like PATHEXT to tell you
        # what file suffixes are executable, so just pass on cmd as-is.
        files = [cmd]

    seen = set()
    for dir in path:
        normdir = os.path.normcase(dir)
        if normdir not in seen:
            seen.add(normdir)
            for thefile in files:
                name = os.path.join(dir, thefile)
                if _access_check(name, mode):
                    return name
    return None


if not is_py3:
    # Simplified backport of Python 3 weakref.finalize:
    # https://github.com/python/cpython/blob/ded4737989316653469763230036b04513cb62b3/Lib/weakref.py#L502-L662
    class finalize(object):
        """Class for finalization of weakrefable objects.

        finalize(obj, func, *args, **kwargs) returns a callable finalizer
        object which will be called when obj is garbage collected. The
        first time the finalizer is called it evaluates func(*arg, **kwargs)
        and returns the result. After this the finalizer is dead, and
        calling it just returns None.

        When the program exits any remaining finalizers will be run.
        """

        # Finalizer objects don't have any state of their own.
        # This ensures that they cannot be part of a ref-cycle.
        __slots__ = ()
        _registry = {}

        def __init__(self, obj, func, *args, **kwargs):
            info = functools.partial(func, *args, **kwargs)
            info.weakref = weakref.ref(obj, self)
            self._registry[self] = info

        def __call__(self):
            """Return func(*args, **kwargs) if alive."""
            info = self._registry.pop(self, None)
            if info:
                return info()

        @classmethod
        def _exitfunc(cls):
            if not cls._registry:
                return
            for finalizer in list(cls._registry):
                try:
                    finalizer()
                except Exception:
                    sys.excepthook(*sys.exc_info())
                assert finalizer not in cls._registry

    atexit.register(finalize._exitfunc)
    weakref.finalize = finalize



"""
Contexts are the "values" that Python would return. However Contexts are at the
same time also the "contexts" that a user is currently sitting in.

A ContextSet is typically used to specify the return of a function or any other
static analysis operation. In jedi there are always multiple returns and not
just one.
"""
from functools import reduce
from operator import add
from parso.python.tree import ExprStmt, SyncCompFor

from rp.libs.jedi import debug
from rp.libs.jedi._compatibility import zip_longest, unicode
from rp.libs.jedi.parser_utils import clean_scope_docstring
from rp.libs.jedi.common import BaseContextSet, BaseContext
from rp.libs.jedi.evaluate.helpers import SimpleGetItemNotFound
from rp.libs.jedi.evaluate.utils import safe_property
from rp.libs.jedi.evaluate.cache import evaluator_as_method_param_cache
from rp.libs.jedi.cache import memoize_method

_sentinel = object()


class HelperContextMixin(object):
    def get_root_context(self):
        context = self
        while True:
            if context.parent_context is None:
                return context
            context = context.parent_context

    @classmethod
    @evaluator_as_method_param_cache()
    def create_cached(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def execute(self, arguments):
        return self.evaluator.execute(self, arguments=arguments)

    def execute_evaluated(self, *value_list):
        from rp.libs.jedi.evaluate.arguments import ValuesArguments
        arguments = ValuesArguments([ContextSet([value]) for value in value_list])
        return self.evaluator.execute(self, arguments)

    def execute_annotation(self):
        return self.execute_evaluated()

    def gather_annotation_classes(self):
        return ContextSet([self])

    def merge_types_of_iterate(self, contextualized_node=None, is_async=False):
        return ContextSet.from_sets(
            lazy_context.infer()
            for lazy_context in self.iterate(contextualized_node, is_async)
        )

    def py__getattribute__(self, name_or_str, name_context=None, position=None,
                           search_global=False, is_goto=False,
                           analysis_errors=True):
        """
        :param position: Position of the last statement -> tuple of line, column
        """
        if name_context is None:
            name_context = self
        from rp.libs.jedi.evaluate import finder
        f = finder.NameFinder(self.evaluator, self, name_context, name_or_str,
                              position, analysis_errors=analysis_errors)
        filters = f.get_filters(search_global)
        if is_goto:
            return f.filter_name(filters)
        return f.find(filters, attribute_lookup=not search_global)

    def py__await__(self):
        await_context_set = self.py__getattribute__(u"__await__")
        if not await_context_set:
            debug.warning('Tried to run __await__ on context %s', self)
        return await_context_set.execute_evaluated()

    def eval_node(self, node):
        return self.evaluator.eval_element(self, node)

    def create_context(self, node, node_is_context=False, node_is_object=False):
        return self.evaluator.create_context(self, node, node_is_context, node_is_object)

    def iterate(self, contextualized_node=None, is_async=False):
        debug.dbg('iterate %s', self)
        if is_async:
            from rp.libs.jedi.evaluate.lazy_context import LazyKnownContexts
            # TODO if no __aiter__ contexts are there, error should be:
            # TypeError: 'async for' requires an object with __aiter__ method, got int
            return iter([
                LazyKnownContexts(
                    self.py__getattribute__('__aiter__').execute_evaluated()
                        .py__getattribute__('__anext__').execute_evaluated()
                        .py__getattribute__('__await__').execute_evaluated()
                        .py__stop_iteration_returns()
                )  # noqa
            ])
        return self.py__iter__(contextualized_node)

    def is_sub_class_of(self, class_context):
        for cls in self.py__mro__():
            if cls.is_same_class(class_context):
                return True
        return False

    def is_same_class(self, class2):
        # Class matching should prefer comparisons that are not this function.
        if type(class2).is_same_class != HelperContextMixin.is_same_class:
            return class2.is_same_class(self)
        return self == class2


class Context(HelperContextMixin, BaseContext):
    """
    Should be defined, otherwise the API returns empty types.
    """
    predefined_names = {}
    """
    To be defined by subclasses.
    """
    tree_node = None

    @property
    def api_type(self):
        # By default just lower name of the class. Can and should be
        # overwritten.
        return self.__class__.__name__.lower()

    def py__getitem__(self, index_context_set, contextualized_node):
        from rp.libs.jedi.evaluate import analysis
        # TODO this context is probably not right.
        analysis.add(
            contextualized_node.context,
            'type-error-not-subscriptable',
            contextualized_node.node,
            message="TypeError: '%s' object is not subscriptable" % self
        )
        return NO_CONTEXTS

    def py__iter__(self, contextualized_node=None):
        if contextualized_node is not None:
            from rp.libs.jedi.evaluate import analysis
            analysis.add(
                contextualized_node.context,
                'type-error-not-iterable',
                contextualized_node.node,
                message="TypeError: '%s' object is not iterable" % self)
        return iter([])

    def get_signatures(self):
        return []

    def is_class(self):
        return False

    def is_instance(self):
        return False

    def is_function(self):
        return False

    def is_module(self):
        return False

    def is_namespace(self):
        return False

    def is_compiled(self):
        return False

    def is_bound_method(self):
        return False

    def py__bool__(self):
        """
        Since Wrapper is a super class for classes, functions and modules,
        the return value will always be true.
        """
        return True

    def py__doc__(self):
        try:
            self.tree_node.get_doc_node
        except AttributeError:
            return ''
        else:
            return clean_scope_docstring(self.tree_node)
        return None

    def get_safe_value(self, default=_sentinel):
        if default is _sentinel:
            raise ValueError("There exists no safe value for context %s" % self)
        return default

    def py__call__(self, arguments):
        debug.warning("no execution possible %s", self)
        return NO_CONTEXTS

    def py__stop_iteration_returns(self):
        debug.warning("Not possible to return the stop iterations of %s", self)
        return NO_CONTEXTS

    def get_qualified_names(self):
        # Returns Optional[Tuple[str, ...]]
        return None

    def is_stub(self):
        # The root context knows if it's a stub or not.
        return self.parent_context.is_stub()


def iterate_contexts(contexts, contextualized_node=None, is_async=False):
    """
    Calls `iterate`, on all contexts but ignores the ordering and just returns
    all contexts that the iterate functions yield.
    """
    return ContextSet.from_sets(
        lazy_context.infer()
        for lazy_context in contexts.iterate(contextualized_node, is_async=is_async)
    )


class _ContextWrapperBase(HelperContextMixin):
    predefined_names = {}

    @safe_property
    def name(self):
        from rp.libs.jedi.evaluate.names import ContextName
        wrapped_name = self._wrapped_context.name
        if wrapped_name.tree_name is not None:
            return ContextName(self, wrapped_name.tree_name)
        else:
            from rp.libs.jedi.evaluate.compiled import CompiledContextName
            return CompiledContextName(self, wrapped_name.string_name)

    @classmethod
    @evaluator_as_method_param_cache()
    def create_cached(cls, evaluator, *args, **kwargs):
        return cls(*args, **kwargs)

    def __getattr__(self, name):
        assert name != '_wrapped_context', 'Problem with _get_wrapped_context'
        return getattr(self._wrapped_context, name)


class LazyContextWrapper(_ContextWrapperBase):
    @safe_property
    @memoize_method
    def _wrapped_context(self):
        with debug.increase_indent_cm('Resolve lazy context wrapper'):
            return self._get_wrapped_context()

    def __repr__(self):
        return '<%s>' % (self.__class__.__name__)

    def _get_wrapped_context(self):
        raise NotImplementedError


class ContextWrapper(_ContextWrapperBase):
    def __init__(self, wrapped_context):
        self._wrapped_context = wrapped_context

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._wrapped_context)


class TreeContext(Context):
    def __init__(self, evaluator, parent_context, tree_node):
        super(TreeContext, self).__init__(evaluator, parent_context)
        self.predefined_names = {}
        self.tree_node = tree_node

    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self.tree_node)


class ContextualizedNode(object):
    def __init__(self, context, node):
        self.context = context
        self.node = node

    def get_root_context(self):
        return self.context.get_root_context()

    def infer(self):
        return self.context.eval_node(self.node)

    def __repr__(self):
        return '<%s: %s in %s>' % (self.__class__.__name__, self.node, self.context)


class ContextualizedName(ContextualizedNode):
    # TODO merge with TreeNameDefinition?!
    @property
    def name(self):
        return self.node

    def assignment_indexes(self):
        """
        Returns an array of tuple(int, node) of the indexes that are used in
        tuple assignments.

        For example if the name is ``y`` in the following code::

            x, (y, z) = 2, ''

        would result in ``[(1, xyz_node), (0, yz_node)]``.

        When searching for b in the case ``a, *b, c = [...]`` it will return::

            [(slice(1, -1), abc_node)]
        """
        indexes = []
        is_star_expr = False
        node = self.node.parent
        compare = self.node
        while node is not None:
            if node.type in ('testlist', 'testlist_comp', 'testlist_star_expr', 'exprlist'):
                for i, child in enumerate(node.children):
                    if child == compare:
                        index = int(i / 2)
                        if is_star_expr:
                            from_end = int((len(node.children) - i) / 2)
                            index = slice(index, -from_end)
                        indexes.insert(0, (index, node))
                        break
                else:
                    raise LookupError("Couldn't find the assignment.")
                is_star_expr = False
            elif node.type == 'star_expr':
                is_star_expr = True
            elif isinstance(node, (ExprStmt, SyncCompFor)):
                break

            compare = node
            node = node.parent
        return indexes


def _getitem(context, index_contexts, contextualized_node):
    from rp.libs.jedi.evaluate.context.iterable import Slice

    # The actual getitem call.
    simple_getitem = getattr(context, 'py__simple_getitem__', None)

    result = NO_CONTEXTS
    unused_contexts = set()
    for index_context in index_contexts:
        if simple_getitem is not None:
            index = index_context
            if isinstance(index_context, Slice):
                index = index.obj

            try:
                method = index.get_safe_value
            except AttributeError:
                pass
            else:
                index = method(default=None)

            if type(index) in (float, int, str, unicode, slice, bytes):
                try:
                    result |= simple_getitem(index)
                    continue
                except SimpleGetItemNotFound:
                    pass

        unused_contexts.add(index_context)

    # The index was somehow not good enough or simply a wrong type.
    # Therefore we now iterate through all the contexts and just take
    # all results.
    if unused_contexts or not index_contexts:
        result |= context.py__getitem__(
            ContextSet(unused_contexts),
            contextualized_node
        )
    debug.dbg('py__getitem__ result: %s', result)
    return result


class ContextSet(BaseContextSet):
    def py__class__(self):
        return ContextSet(c.py__class__() for c in self._set)

    def iterate(self, contextualized_node=None, is_async=False):
        from rp.libs.jedi.evaluate.lazy_context import get_merged_lazy_context
        type_iters = [c.iterate(contextualized_node, is_async=is_async) for c in self._set]
        for lazy_contexts in zip_longest(*type_iters):
            yield get_merged_lazy_context(
                [l for l in lazy_contexts if l is not None]
            )

    def execute(self, arguments):
        return ContextSet.from_sets(c.evaluator.execute(c, arguments) for c in self._set)

    def execute_evaluated(self, *args, **kwargs):
        return ContextSet.from_sets(c.execute_evaluated(*args, **kwargs) for c in self._set)

    def py__getattribute__(self, *args, **kwargs):
        if kwargs.get('is_goto'):
            return reduce(add, [c.py__getattribute__(*args, **kwargs) for c in self._set], [])
        return ContextSet.from_sets(c.py__getattribute__(*args, **kwargs) for c in self._set)

    def get_item(self, *args, **kwargs):
        return ContextSet.from_sets(_getitem(c, *args, **kwargs) for c in self._set)

    def try_merge(self, function_name):
        context_set = self.__class__([])
        for c in self._set:
            try:
                method = getattr(c, function_name)
            except AttributeError:
                pass
            else:
                context_set |= method()
        return context_set

    def gather_annotation_classes(self):
        return ContextSet.from_sets([c.gather_annotation_classes() for c in self._set])

    def get_signatures(self):
        return [sig for c in self._set for sig in c.get_signatures()]


NO_CONTEXTS = ContextSet([])


def iterator_to_context_set(func):
    def wrapper(*args, **kwargs):
        return ContextSet(func(*args, **kwargs))

    return wrapper



def header_quopri_check(c): ...
def body_quopri_check(c): ...
def header_quopri_len(s): ...
def body_quopri_len(str): ...
def unquote(s): ...
def quote(c): ...
def header_encode(header, charset: str = ..., keep_eols: bool = ..., maxlinelen: int = ..., eol=...): ...
def encode(body, binary: bool = ..., maxlinelen: int = ..., eol=...): ...

body_encode = encode
encodestring = encode

def decode(encoded, eol=...): ...

body_decode = decode
decodestring = decode

def header_decode(s): ...



copy=paste=None
try:
    """
    Pyperclip
    
    A cross-platform clipboard module for Python. (only handles plain text for now)
    By Al Sweigart al@inventwithpython.com
    BSD License
    
    Usage:
      import pyperclip
      pyperclip.copy('The text to be copied to the clipboard.')
      spam = pyperclip.paste()
    
    On Windows, no additional modules are needed.
    On Mac, this module makes use of the pbcopy and pbpaste commands, which should come with the os.
    On Linux, this module makes use of the xclip or xsel commands, which should come with the os. Otherwise run "sudo apt-get install xclip" or "sudo apt-get install xsel"
      Otherwise on Linux, you will need the gtk or PyQt4 modules installed.
    
    The gtk module is not available for Python 3, and this module does not work with PyGObject yet.
    """

    __version__ = '1.5.6'

    import platform, os
    from subprocess import call, Popen, PIPE


    def _pasteWindows():
        CF_UNICODETEXT = 13
        d = ctypes.windll
        d.user32.OpenClipboard(None)
        handle = d.user32.GetClipboardData(CF_UNICODETEXT)
        data = ctypes.c_wchar_p(handle).value
        d.user32.CloseClipboard()
        return data


    def _copyWindows(text):
        GMEM_DDESHARE = 0x2000
        CF_UNICODETEXT = 13
        d = ctypes.windll # cdll expects 4 more bytes in user32.OpenClipboard(None)
        try:  # Python 2
            if not isinstance(text, unicode):
                text = text.decode('mbcs')
        except NameError:
            if not isinstance(text, str):
                text = text.decode('mbcs')
        d.user32.OpenClipboard(None)
        d.user32.EmptyClipboard()
        hCd = d.kernel32.GlobalAlloc(GMEM_DDESHARE, len(text.encode('utf-16-le')) + 2)
        pchData = d.kernel32.GlobalLock(hCd)
        ctypes.cdll.msvcrt.wcscpy(ctypes.c_wchar_p(pchData), text)
        d.kernel32.GlobalUnlock(hCd)
        d.user32.SetClipboardData(CF_UNICODETEXT, hCd)
        d.user32.CloseClipboard()


    def _pasteCygwin():
        CF_UNICODETEXT = 13
        d = ctypes.cdll
        d.user32.OpenClipboard(None)
        handle = d.user32.GetClipboardData(CF_UNICODETEXT)
        data = ctypes.c_wchar_p(handle).value
        d.user32.CloseClipboard()
        return data


    def _copyCygwin(text):
        GMEM_DDESHARE = 0x2000
        CF_UNICODETEXT = 13
        d = ctypes.cdll
        try:  # Python 2
            if not isinstance(text, unicode):
                text = text.decode('mbcs')
        except NameError:
            if not isinstance(text, str):
                text = text.decode('mbcs')
        d.user32.OpenClipboard(None)
        d.user32.EmptyClipboard()
        hCd = d.kernel32.GlobalAlloc(GMEM_DDESHARE, len(text.encode('utf-16-le')) + 2)
        pchData = d.kernel32.GlobalLock(hCd)
        ctypes.cdll.msvcrt.wcscpy(ctypes.c_wchar_p(pchData), text)
        d.kernel32.GlobalUnlock(hCd)
        d.user32.SetClipboardData(CF_UNICODETEXT, hCd)
        d.user32.CloseClipboard()


    def _copyOSX(text):
        text = str(text)
        p = Popen(['pbcopy', 'w'], stdin=PIPE)
        try:
            # works on Python 3 (bytes() requires an encoding)
            p.communicate(input=bytes(text, 'utf-8'))
        except TypeError:
            # works on Python 2 (bytes() only takes one argument)
            p.communicate(input=bytes(text))


    def _pasteOSX():
        p = Popen(['pbpaste', 'r'], stdout=PIPE)
        stdout, stderr = p.communicate()
        return bytes.decode(stdout)


    def _pasteGtk():
        return gtk.Clipboard().wait_for_text()


    def _copyGtk(text):
        global cb
        text = str(text)
        cb = gtk.Clipboard()
        cb.set_text(text)
        cb.store()


    def _pasteQt():
        return str(cb.text())


    def _copyQt(text):
        text = str(text)
        cb.setText(text)


    def _copyXclip(text):
        p = Popen(['xclip', '-selection', 'c'], stdin=PIPE)
        try:
            # works on Python 3 (bytes() requires an encoding)
            p.communicate(input=bytes(text, 'utf-8'))
        except TypeError:
            # works on Python 2 (bytes() only takes one argument)
            p.communicate(input=bytes(text))


    def _pasteXclip():
        p = Popen(['xclip', '-selection', 'c', '-o'], stdout=PIPE)
        stdout, stderr = p.communicate()
        return bytes.decode(stdout)


    def _copyXsel(text):
        p = Popen(['xsel', '-i'], stdin=PIPE)
        try:
            # works on Python 3 (bytes() requires an encoding)
            p.communicate(input=bytes(text, 'utf-8'))
        except TypeError:
            # works on Python 2 (bytes() only takes one argument)
            p.communicate(input=bytes(text))


    def _pasteXsel():
        p = Popen(['xsel', '-o'], stdout=PIPE)
        stdout, stderr = p.communicate()
        return bytes.decode(stdout)



    # Determine the OS/platform and set the copy() and paste() functions accordingly.
    if 'cygwin' in platform.system().lower():
        _functions = 'Cygwin' # for debugging
        import ctypes
        paste = _pasteCygwin
        copy = _copyCygwin
    elif os.name == 'nt' or platform.system() == 'Windows':
        _functions = 'Windows' # for debugging
        import ctypes
        paste = _pasteWindows
        copy = _copyWindows
    elif os.name == 'mac' or platform.system() == 'Darwin':
        _functions = 'OS X pbcopy/pbpaste' # for debugging
        paste = _pasteOSX
        copy = _copyOSX
    elif os.name == 'posix' or platform.system() == 'Linux':
        # Determine which command/module is installed, if any.
        xclipExists = call(['which', 'xclip'],
                           stdout=PIPE, stderr=PIPE) == 0

        xselExists = call(['which', 'xsel'],
                          stdout=PIPE, stderr=PIPE) == 0

        gtkInstalled = False
        try:
            # Check it gtk is installed.
            import gtk
            gtkInstalled = True
        except ImportError:
            pass

        if not gtkInstalled:
            # Check if PyQt4 is installed.
            PyQt4Installed = False
            try:
                import PyQt4.QtCore
                import PyQt4.QtGui
                PyQt4Installed = True
            except ImportError:
                pass

        # Set one of the copy & paste functions.
        if xclipExists:
            _functions = 'xclip command' # for debugging
            paste = _pasteXclip
            copy = _copyXclip
        elif gtkInstalled:
            _functions = 'gtk module' # for debugging
            paste = _pasteGtk
            copy = _copyGtk
        elif PyQt4Installed:
            _functions = 'PyQt4 module' # for debugging
            app = PyQt4.QtGui.QApplication([])
            cb = PyQt4.QtGui.QApplication.clipboard()
            paste = _pasteQt
            copy = _copyQt
        elif xselExists:
            # TODO: xsel doesn't seem to work on Raspberry Pi (my test Linux environment). Putting this as the last method tried.
            _functions = 'xsel command' # for debugging
            paste = _pasteXsel
            copy = _copyXsel
        else:
            raise Exception('Pyperclip requires the xclip or xsel application, or the gtk or PyQt4 module.')
    else:
        raise RuntimeError('pyperclip does not support your system.')
except:
    print("Pyperclip disabled")


from google.protobuf.any_pb2 import (
    Any,
)
from google.protobuf.internal.containers import (
    RepeatedCompositeFieldContainer,
)
from google.protobuf.message import (
    Message,
)
from typing import (
    Iterable,
    Optional,
)


class TestAny(Message):
    int32_value: int

    @property
    def any_value(self) -> Any: ...

    @property
    def repeated_any_value(self) -> RepeatedCompositeFieldContainer[Any]: ...

    def __init__(self,
                 int32_value: Optional[int] = ...,
                 any_value: Optional[Any] = ...,
                 repeated_any_value: Optional[Iterable[Any]] = ...,
                 ) -> None: ...

    @classmethod
    def FromString(cls, s: bytes) -> TestAny: ...



# Stubs for pprint

# Based on http://docs.python.org/2/library/pprint.html
# Based on http://docs.python.org/3/library/pprint.html

import sys
from typing import Any, Dict, Tuple, IO

if sys.version_info >= (3, 4):
    def pformat(o: object, indent: int = ..., width: int = ...,
                depth: int = ..., compact: bool = ...) -> str: ...
else:
    def pformat(o: object, indent: int = ..., width: int = ...,
                depth: int = ...) -> str: ...

if sys.version_info >= (3, 4):
    def pprint(o: object, stream: IO[str] = ..., indent: int = ..., width: int = ...,
               depth: int = ..., compact: bool = ...) -> None: ...
else:
    def pprint(o: object, stream: IO[str] = ..., indent: int = ..., width: int = ...,
               depth: int = ...) -> None: ...

def isreadable(o: object) -> bool: ...
def isrecursive(o: object) -> bool: ...
def saferepr(o: object) -> str: ...

class PrettyPrinter:
    if sys.version_info >= (3, 4):
        def __init__(self, indent: int = ..., width: int = ..., depth: int = ...,
                     stream: IO[str] = ..., compact: bool = ...) -> None: ...
    else:
        def __init__(self, indent: int = ..., width: int = ..., depth: int = ...,
                     stream: IO[str] = ...) -> None: ...

    def pformat(self, o: object) -> str: ...
    def pprint(self, o: object) -> None: ...
    def isreadable(self, o: object) -> bool: ...
    def isrecursive(self, o: object) -> bool: ...
    def format(self, o: object, context: Dict[int, Any], maxlevels: int,
               level: int) -> Tuple[str, bool, bool]: ...
